'''
This file appears to be an old version of what is now ConsRepAgentModel.py.
Its previous contents have been entirely removed and replaced with a universal
import from ConsRepAgentModel.  Whenever a user imports from this file, they
will get a warning that they should import from ConsRepAgentModel instead.
'''

import warnings
from HARK.ConsumptionSaving.ConsRepAgentModel import *

warnings.warn('Please import from ConsRepAgentModel rather than RepAgentModel.  This module will be removed in a future version of HARK.')